# -*- coding: utf-8
from __future__ import unicode_literals

import sys

GREY    = 30
RED     = 31
GREEN   = 32
YELLOW  = 33
BLUE    = 34
MAGENTA = 35
CYAN    = 36
WHITE   = 37
BRIGHT_WHITE = 57
NORMAL = 0
BOLD = 1


class ColoredString(type('')):
    color = WHITE
    weight = NORMAL
    def __str__(self):
        if sys.stdout.isatty():
            return '\033[%sm\033[%sm%s\033[0m' % (self.weight, self.color, super(ColoredString, self).__str__())
        else:
            return super(ColoredString, self).__str__()


class Added(ColoredString):
    color = GREEN

class Deleted(ColoredString):
    color = RED

class Modified(ColoredString):
    color = YELLOW

class LineNumber(ColoredString):
    color = MAGENTA

class Unknown(ColoredString):
    color = MAGENTA

class Error(ColoredString):
    color = GREY

class DeletedHeader(Deleted):
    weight = BOLD

class AddedHeader(Added):
    weight = BOLD

class Header(ColoredString):
    color = BRIGHT_WHITE
    weight = BOLD

class Prompt(ColoredString):
    color = YELLOW
    weight = BOLD

    def __new__(cls, s):
        return super(Prompt, cls).__new__(cls, s+' ')



# -*- coding: utf-8
from __future__ import unicode_literals

import os
import sys
import hashlib
import argparse
#import difflib

try:
    from itertools import izip as zip
except ImportError:
    pass

from . import __version__, FragmentsError, _iterate_over_files, _smart_open
from .config import FragmentsConfig, configuration_directory_name, find_configuration, ConfigurationFileCorrupt, ConfigurationFileNotFound, ConfigurationDirectoryNotFound
from .diff import _full_diff
from .apply import apply
from .precisecodevillemerge import Weave
from . import color

class ExecutionError(FragmentsError): pass

def help(*args):
    """Prints help."""
    from . import commands
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, help.__name__), description=help.__doc__)
    parser.add_argument('COMMAND', help="command to show help for", nargs="?", choices=__all__)
    args = parser.parse_args(args)
    if args.COMMAND:
        for l in getattr(commands, args.COMMAND)('-h'):
            yield l
    else:
        parser.parse_args(['-h'])


def init(*args):
    """Initialize a new fragments repository. Repository is in a directory named _fragments/, created in either the current working directory, or FRAGMENTS_ROOT if specified."""
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, init.__name__), description=init.__doc__)
    parser.add_argument('FRAGMENTS_ROOT', help="root directory in which to create the _fragments/ directory", nargs="?")
    args = parser.parse_args(args)

    try:
        config = FragmentsConfig()
    except ConfigurationFileCorrupt as exc:
        config = FragmentsConfig(autoload=False)
        os.rename(config.path, config.path + '.corrupt')
        config.dump()
    except ConfigurationFileNotFound as exc:
        config = FragmentsConfig(autoload=False)
        config.dump()
    except ConfigurationDirectoryNotFound as exc:
        if args.FRAGMENTS_ROOT:
            configuration_parent = os.path.realpath(args.FRAGMENTS_ROOT)
        else:
            configuration_parent = os.path.split(os.getcwd())[0]
        if os.access(configuration_parent, os.R_OK|os.W_OK):
            configuration_path = os.path.join(configuration_parent, configuration_directory_name)
            os.mkdir(configuration_path)
            config = FragmentsConfig(configuration_path, autoload=False)
            config.dump()
        else:
            raise ExecutionError("Could not create fragments directory in '%s', aborting.\n(Do you have the correct permissions?)" % configuration_parent)
    else:
        raise ExecutionError("Current fragments configuration found in '%s', aborting." % config.path)
    yield "Fragments configuration created in '%s'" % config.path


def _file_status(config, curr_path):
    key = curr_path[len(config.root)+1:]
    if key not in config['files']:
        return '?' # unfollowed

    repo_path = os.path.join(config.directory, config['files'][key])

    repo_exists = os.access(repo_path, os.R_OK|os.W_OK)
    curr_exists = os.access(curr_path, os.R_OK|os.W_OK)

    if repo_exists and curr_exists:
        if os.stat(repo_path)[6] != os.stat(curr_path)[6]:
            return 'M' # current and repo versions have different sizes: file has been modified
        else:
            for repo_line, curr_line in zip(open(repo_path, 'r').readlines(), open(curr_path, 'r').readlines()):
                if len(repo_line) != len(curr_line):
                    return 'M' # corresponding lines have different length: file has been modified
                if repo_line != curr_line:
                    return 'M' # corresponding lines are different: file has been modified
            return ' ' # current and repo versions are the same size, corresponding lines are all the same length and all match: file is unmodified
    elif repo_exists:
        return 'D' # deleted
    elif curr_exists:
        return 'A' # added
    else:
        return 'E' # error. this should never happen - both files on disk are missing, but file is being followed


_status_to_color = {
    '?':color.Unknown,
    'M':color.Modified,
    'D':color.Deleted,
    'A':color.Added,
    'E':color.Error,
}


def status(*args):
    """
    Get the current status of the fragments repository, limited to FILENAME(s) if specified.
    Limit output to files with status STATUS, if present.
    """
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, status.__name__), description=status.__doc__)
    parser.add_argument('FILENAME', help="files to show status for", nargs="*")
    parser.add_argument('-l', '--limit', type=str, dest="STATUS", default=None, action="store", help="limit to files in STATUS")
    args = parser.parse_args(args)

    config = FragmentsConfig()
    if not args.STATUS:
        yield "%s configuration version %s.%s.%s" % ((__package__,) + config['version'])
        yield "stored in %s" % config.directory
    for curr_path in _iterate_over_files(args.FILENAME, config):
        s = _file_status(config, curr_path)
        if not args.STATUS or s in args.STATUS.upper():
            yield _status_to_color.get(s, str)('%s\t%s' % (s, os.path.relpath(curr_path)))


def follow(*args):
    """Start following changes to one or more FILENAME(s)."""
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, follow.__name__), description=follow.__doc__)
    parser.add_argument('FILENAME', help="files to follow", nargs="+")
    args = parser.parse_args(args)

    config = FragmentsConfig()
    for filename in set(args.FILENAME):
        fullpath = os.path.realpath(filename)
        if fullpath.startswith(config.root):
            key = os.path.relpath(fullpath, config.root)
            if key in config['files']:
                yield "'%s' is already being followed" % os.path.relpath(filename)
                continue
            if os.access(fullpath, os.W_OK|os.R_OK):
                file_sha = hashlib.sha256(('%s:%s' % (__package__, key)).encode('utf8')).hexdigest()
                config['files'][key] = file_sha
                yield "'%s' is now being followed (SHA-256: '%s')" % (os.path.relpath(filename), file_sha)
            else:
                yield "Could not access '%s' to follow it" % os.path.relpath(filename)
        else:
            yield "Could not follow '%s'; it is outside the repository" % os.path.relpath(filename)
    config.dump()


def forget(*args):
    """Stop following changes to one or more FILENAME(s)."""
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, forget.__name__), description=forget.__doc__)
    parser.add_argument('FILENAME', help="files to forget", nargs="+")
    args = parser.parse_args(args)

    config = FragmentsConfig()
    for filename in set(args.FILENAME):
        fullpath = os.path.realpath(filename)
        if fullpath.startswith(config.root):
            key = os.path.relpath(fullpath, config.root)
            if key in config['files']:
                file_sha = config['files'][key]
                sha_path = os.path.join(config.directory, file_sha)
                if os.access(os.path.join(config.directory, file_sha), os.W_OK|os.R_OK):
                    os.unlink(sha_path)
                    yield "'%s' is no longer being followed" % os.path.relpath(filename)
                else:
                    yield "'%s' was never committed and will not be followed" % os.path.relpath(filename)
                del config['files'][key]
            else:
                yield "Could not forget '%s', it was not being followed" % os.path.relpath(filename)
        else:
            yield "Could not forget '%s'; it is outside the repository" % os.path.relpath(filename)
    config.dump()


def rename(*args):
    """Rename OLD_FILENAME to NEW_FILENAME, moving the actual file on disk if it has not already been moved."""
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, rename.__name__), description=rename.__doc__)
    parser.add_argument('OLD_FILENAME', help="old file name")
    parser.add_argument('NEW_FILENAME', help="new file name")
    args = parser.parse_args(args)

    old_name, new_name = os.path.relpath(args.OLD_FILENAME), os.path.relpath(args.NEW_FILENAME)

    config = FragmentsConfig()
    old_path = os.path.realpath(old_name)
    old_key = os.path.relpath(old_path, config.root)
    new_path = os.path.realpath(new_name)
    new_key = os.path.relpath(new_path, config.root)
    if old_key not in config['files']:
        yield "Could not rename '%s', it is not being tracked" % old_name
    elif new_key in config['files']:
        yield "Could not rename '%s' to '%s', '%s' is already being tracked" % (old_name, new_name, new_name)
    elif os.access(old_path, os.W_OK|os.R_OK) and os.access(new_path, os.W_OK|os.R_OK):
        yield "Could not rename '%s' to '%s', both files already exist" % (old_name, new_name)
    elif not os.access(old_path, os.W_OK|os.R_OK) and not os.access(new_path, os.W_OK|os.R_OK):
        yield "Could not rename '%s' to '%s', neither file exists" % (old_name, new_name)
    else:
        config['files'][new_key] = config['files'][old_key]
        del config['files'][old_key]
        if os.access(old_path, os.W_OK|os.R_OK):
            os.rename(old_path, new_path)
    config.dump()


def diff(*args):
    """Show differences between committed and uncommitted versions, limited to FILENAME(s) if specified."""
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, diff.__name__), description=diff.__doc__)
    parser.add_argument('FILENAME', help="file(s) to show changes in", nargs="*")
    parser.add_argument('-U', '--unified', type=int, dest="NUM", default=3, action="store", help="number of lines of context to show")
    args = parser.parse_args(args)

    config = FragmentsConfig()

    for curr_path in _iterate_over_files(args.FILENAME, config):
        key = os.path.relpath(curr_path, config.root)
        if key not in config['files']:
            yield "Could not diff '%s', it is not being followed" % os.path.relpath(curr_path)
            continue

        s = _file_status(config, curr_path)
        if s in 'MAD':
            repo_lines = []
            curr_lines = []
            if s in 'MA':
                curr_lines = _smart_open(curr_path, 'r').readlines()
            if s in 'MD':
                repo_lines = _smart_open(os.path.join(config.directory, config['files'][key]), 'r').readlines()
            weave = Weave()
            weave.add_revision(1, repo_lines, [])
            weave.add_revision(2, curr_lines, [])
            for l in _full_diff(weave.merge(1, 2), key, context_lines=args.NUM):
                yield l
            # for dl in difflib.unified_diff(repo_lines, curr_lines, fromfile=key, tofile=key):
            #     yield dl


def commit(*args):
    """Commit changes to the fragments repository, limited to FILENAME(s) if specified."""
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, commit.__name__), description=commit.__doc__)
    parser.add_argument('FILENAME', help="file(s) to commit", nargs="*")
    args = parser.parse_args(args)

    config = FragmentsConfig()

    for curr_path in _iterate_over_files(args.FILENAME, config):
        key = os.path.relpath(curr_path, config.root)
        if key not in config['files']:
            yield "Could not commit '%s' because it is not being followed" % os.path.relpath(curr_path)
            continue

        s = _file_status(config, curr_path)
        if s in 'MA':
            repo_path = os.path.join(config.directory, config['files'][key])
            with _smart_open(repo_path, 'w') as repo_file:
                repo_file.write(_smart_open(curr_path, 'r').read())
            os.utime(repo_path, os.stat(curr_path)[7:9])
            yield "'%s' committed" % os.path.relpath(curr_path)
        elif s == 'D':
            yield "Could not commit '%s' because it has been removed, instead revert or forget it" % os.path.relpath(curr_path)
        elif s == ' ':
            yield "Could not commit '%s' because it has not been changed" % os.path.relpath(curr_path)


def revert(*args):
    """Revert changes to the fragments repository, limited to FILENAME(s) if specified."""
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, revert.__name__), description=revert.__doc__)
    parser.add_argument('FILENAME', help="file(s) to revert", nargs="*")
    args = parser.parse_args(args)

    config = FragmentsConfig()

    for curr_path in _iterate_over_files(args.FILENAME, config):
        key = os.path.relpath(curr_path, config.root)
        if key not in config['files']:
            yield "Could not revert '%s' because it is not being followed" % os.path.relpath(curr_path)
            continue

        s = _file_status(config, curr_path)
        if s in 'MD':
            repo_path = os.path.join(config.directory,  config['files'][key])
            with _smart_open(curr_path, 'w') as curr_file:
                curr_file.write(_smart_open(repo_path, 'r').read())
            os.utime(curr_path, os.stat(repo_path)[7:9])
            yield "'%s' reverted" % key
        elif s == 'A':
            yield "Could not revert '%s' because it has never been committed" % os.path.relpath(curr_path)
        elif s == ' ':
            yield "Could not revert '%s' because it has not been changed" % os.path.relpath(curr_path)


def fork(*args):
    """
    Create a new file in TARGET_FILENAME based on one or more SOURCE_FILENAME(s).
    Large common sections are preserved;
    differing sections, and common sections shorter than NUM lines between differing sections, are replaced with one newline for each line or conflict.
    """
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, fork.__name__), description=fork.__doc__)
    parser.add_argument('SOURCE_FILENAME', help="old file names", nargs="+")
    parser.add_argument('TARGET_FILENAME', help="new file name")
    parser.add_argument('-U', '--unified', type=int, dest="NUM", default=3, action="store", help="number of lines of context to use")
    args = parser.parse_args(args)

    config = FragmentsConfig()
    new_path = os.path.realpath(args.TARGET_FILENAME)
    new_key = os.path.relpath(new_path, config.root)
    if new_key in config['files']:
        yield "Could not fork into '%s', it is already followed" % os.path.relpath(args.TARGET_FILENAME)
        return
    if os.access(new_path, os.R_OK|os.W_OK):
        yield "Could not fork into '%s', the file already exists" % os.path.relpath(args.TARGET_FILENAME)
        return

    old_filenames = []
    for old_name in args.SOURCE_FILENAME:
        old_path = os.path.realpath(old_name)
        old_key = os.path.relpath(old_path, config.root)
        if os.access(old_path, os.R_OK|os.W_OK):
            old_filenames.append(old_path)
            if old_key not in config['files']:
                yield "Warning, '%s' not being followed" % os.path.relpath(old_path)
        else:
            yield "Skipping '%s' while forking, it does not exist" % os.path.relpath(old_path)

    if not old_filenames:
        yield "Could not fork; no valid source files specified"
        return

    weave = Weave()

    new_lines = _smart_open(old_filenames[0], 'r').readlines()
    previous_revision = 1
    weave.add_revision(previous_revision, new_lines, [])
    for old_name in old_filenames[1:]:
        current_revision = previous_revision + 1
        weave.add_revision(current_revision, _smart_open(old_name, 'r').readlines(), [])
        new_lines = []
        diff_output = weave.merge(previous_revision, current_revision)
        i = 0
        while i < len(diff_output):
            line_or_conflict = diff_output[i]
            if isinstance(line_or_conflict, tuple):
                old, new = line_or_conflict
                following_conflict_index = 0 # index of furthest following conflict within args.NUM lines
                for j, loc in enumerate(diff_output[i+1:i+1+args.NUM]):
                    if isinstance(loc, tuple):
                        following_conflict_index = j
                new_lines.extend(['\n'] * (following_conflict_index+1)) # add a blank line for each line and conflict we are skipping
                i += following_conflict_index
            else:
                new_lines.append(line_or_conflict)
            i += 1
        previous_revision = current_revision + 1
        weave.add_revision(previous_revision, new_lines, [])

    with _smart_open(new_path, 'w') as new_file:
        new_file.writelines(new_lines)
    yield "Forked new file in '%s', remember to follow and commit it" % os.path.relpath(args.TARGET_FILENAME)
    config.dump()


try:
    raw_input
except NameError: # pragma: no cover # Python 3 support
    raw_input = input

def _main(): # pragma: no cover
    from . import commands
    print("%s version %s.%s.%s" % ((__package__,) + __version__))
    cmd = None
    if len(sys.argv) > 1:
        if sys.argv[1] in __all__:
            cmd = sys.argv[1]
        else:
            cmds = [c for c in __all__ if c.startswith(sys.argv[1])]
            if len(cmds) == 1:
                cmd = cmds.pop()
            else:
                if len(cmds) > 1:
                    print("Command '%s' is ambiguous, did you mean:" % sys.argv[1])
                    print(' '.join(cmds))
                else:
                    print("No such command '%s'. Available commands are:" % sys.argv[1])
                    print(' '.join(__all__))
    if (cmd): # command is present and legit
        try:
            command_generator = getattr(commands, cmd)(*sys.argv[2:])
            while True:
                try:
                    l = next(command_generator)
                    if isinstance(l, color.Prompt):
                        response = raw_input(l)
                        l = command_generator.send(response.strip())
                    print(str(l))
                except StopIteration:
                    break
        except FragmentsError as exc:
            sys.exit(exc.args[0])
        except KeyboardInterrupt:
            pass

__all__ = ['help', 'init', 'status', 'follow', 'forget', 'rename', 'diff', 'commit', 'revert', 'fork', 'apply']



# -*- coding: utf-8
from __future__ import unicode_literals

from . import color

def _visible_in_diff(merge_result, context_lines=3):
    """Collects the set of lines that should be visible in a diff with a certain number of context lines"""
    i = old_line = new_line = 0
    while i < len(merge_result):
        line_or_conflict = merge_result[i]
        if isinstance(line_or_conflict, tuple):
            yield old_line, new_line, line_or_conflict
            old_line += len(line_or_conflict[0])
            new_line += len(line_or_conflict[1])
        else:
            for j in (list(range(max(0, i-context_lines), i                                        )) +  # look behind for nearby conflicts
                      list(range(i+1                    , min(len(merge_result), i+1+context_lines)))):  # look ahead for nearby conflicts
                if isinstance(merge_result[j], tuple):
                    yield old_line, new_line, line_or_conflict
                    break
            else:
                yield None # sentinel to mark boundaries between diff section groups
            old_line += 1
            new_line += 1
        i += 1
    yield None


def _split_diff(merge_result, context_lines=3):
    """Split diffs and context lines into groups based on None sentinel"""
    collect = []
    for item in _visible_in_diff(merge_result, context_lines=context_lines):
        if item is None:
            if collect:
                yield collect
            collect = []
        else:
            collect.append(item)


def _diff_group_position(group):
    """Generate a unified diff position line for a diff group"""
    old_start = group[0][0]
    new_start = group[0][1]
    old_length = new_length = 0
    for old_line, new_line, line_or_conflict in group:
        if isinstance(line_or_conflict, tuple):
            old, new = line_or_conflict
            old_length += len(old)
            new_length += len(new)
        else:
            old_length += 1
            new_length += 1
    if old_length:
        old_start += 1
    if new_length:
        new_start += 1

    return color.LineNumber('@@ -%s,%s +%s,%s @@' % (old_start, old_length, new_start, new_length))


def _diff_group(group):
    """Generate a diff section for diff group"""
    yield _diff_group_position(group)

    for old_line, new_line, line_or_conflict in group:
        if isinstance(line_or_conflict, tuple):
            old, new = line_or_conflict
            for o in old:
                yield color.Deleted('-' + o.strip('\n'))

            if new and old and new[-1].endswith('\n') and not old[-1].endswith('\n'): # new last line has a newline but old last line doesn't
                yield "\ No newline at end of file"

            for n in new:
                yield color.Added('+' + n.strip('\n'))

            if old and new and old[-1].endswith('\n') and not new[-1].endswith('\n'): # old last line has a newline but new last line doesn't
                yield "\ No newline at end of file"

        else:
            yield ' ' + line_or_conflict.strip('\n')


def _full_diff(merge_result, key, context_lines=3):
    """Generate a full diff based on a Weave merge result"""
    header_printed = False
    for group in _split_diff(merge_result, context_lines=context_lines):
        if not header_printed:
            header_printed = True
            yield color.Header('diff a/%s b/%s' % (key, key))
            yield color.DeletedHeader('--- %s' % key)
            yield color.AddedHeader('+++ %s' % key)

        for l in _diff_group(group):
            yield l



# -*- coding: utf-8
from __future__ import unicode_literals

import os
import codecs

__version__ = (1,2,1)

class FragmentsError(Exception): pass

def _iterate_over_files(args, config):
    if args:
        seen = set()
        for a in args:
            if a not in seen:
                yield os.path.realpath(a)
                seen.add(a)
    else:
        for f in sorted(config['files']):
            yield os.path.join(config.root, f)

def _smart_open(path, mode='r'):
    return codecs.open(path, mode=mode, encoding='utf8')



# -*- coding: utf-8
from __future__ import unicode_literals

import os
import sys
import json

from . import FragmentsError, __version__


configuration_file_name = 'config.json'
configuration_directory_name = '_fragments'


class ConfigurationError(FragmentsError): pass
class ConfigurationDirectoryNotFound(ConfigurationError): pass
class ConfigurationFileNotFound(ConfigurationError): pass
class ConfigurationFileCorrupt(ConfigurationError): pass


def find_configuration(current=None):
    current = current or os.getcwd()
    path = current
    while True:
        configuration_path = os.path.join(path, configuration_directory_name)
        if os.path.exists(path) and os.path.exists(configuration_path):
            return configuration_path
        path, oldpath = os.path.split(path)[0], path
        if oldpath == path:
            raise ConfigurationDirectoryNotFound("Could not find fragments configuration directory in %r or any parent directories" % current)


class FragmentsConfig(dict):

    defaults = {
        'files': {},
        'version': __version__,
    }

    def __init__(self, directory=None, autoload=True):
        if directory is None:
            directory = find_configuration()
        self.directory = directory
        self.path = os.path.join(self.directory, configuration_file_name)
        self.root = os.path.split(self.directory)[0]
        self.update(FragmentsConfig.defaults)
        if autoload:
            self.load()

    def load(self):
        if os.access(self.path, os.R_OK|os.W_OK):
            file_contents = open(self.path, 'r').read()
            try:
                parsed_json = json.loads(file_contents)
            except Exception as exc:
                raise ConfigurationFileCorrupt(exc.args[0])
            self.update(parsed_json)
            self['version'] = tuple(self['version'])
        else:
            raise ConfigurationFileNotFound("Could not access %r, if the file exists, check its permissions" % self.path)

    def dump(self):
        self['version'] = __version__
        open(self.path, 'w').write(json.dumps(self, sort_keys=True, indent=4))



# -*- coding: utf-8
from __future__ import unicode_literals

import os
import argparse

from . import _iterate_over_files, _smart_open
from .precisecodevillemerge import Weave
from .config import FragmentsConfig
from .diff import _diff_group, _split_diff
from .color import Prompt


def apply(*args):
    """
    Apply changes in SOURCE_FILENAME that were made since last commit, where possible.
    Limit application to TARGET_FILENAME(s) if specified.
    Files that conflict in their entirety will be skipped.
    Smaller conflicts will be written to the file as conflict sections.

    In interactive mode, you can use the following commands:

        y - include this change
        n - do not include this change
        a - include this change and all remaining changes
        d - done, do not include this change nor any remaining changes
        j - leave this change undecided, see next undecided change
        k - leave this change undecided, see previous undecided change
        ? - interactive apply mode help
    """
    parser = argparse.ArgumentParser(prog="%s %s" % (__package__, apply.__name__), description=apply.__doc__)
    parser.add_argument('SOURCE_FILENAME', help="file containing changes to be applied")
    parser.add_argument('TARGET_FILENAME', help="file(s) to apply changes to", nargs='*')
    parser.add_argument('-U', '--unified', type=int, dest="NUM", default=3, action="store", help="number of lines of context to show")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-i", "--interactive", action="store_true" , default=True , dest="interactive", help="interactively select changes to apply")
    group.add_argument("-a", "--automatic"  , action="store_false", default=False, dest="interactive", help="automatically apply all changes")
    args = parser.parse_args(args)

    config = FragmentsConfig()
    weave = Weave()
    changed_path = os.path.realpath(args.SOURCE_FILENAME)
    changed_key = os.path.relpath(changed_path, config.root)
    if changed_key not in config['files']:
        yield "Could not apply changes in '%s', it is not being followed" % os.path.relpath(changed_path)
        return
    elif not os.access(changed_path, os.R_OK|os.W_OK):
        yield "Could not apply changes in '%s', it no longer exists on disk" % os.path.relpath(changed_path)
        return

    old_path = os.path.join(config.directory, config['files'][changed_key])

    if not os.access(old_path, os.R_OK|os.W_OK):
        yield "Could not apply changes in '%s', it has never been committed" % os.path.relpath(changed_path)
        return

    old_revision = 1
    weave.add_revision(old_revision, _smart_open(old_path, 'r').readlines(), [])
    new_revision = 2
    weave.add_revision(new_revision, _smart_open(changed_path, 'r').readlines(), [])

    diff = weave.merge(old_revision, new_revision)

    # Select the chunks to be applied, possibly interactively
    preserve_changes = {}
    discard_changes = {}
    display_groups = list(_split_diff(diff, context_lines=args.NUM))
    index = 0
    apply_all = None
    while display_groups:
        display_group = display_groups[index]
        if apply_all is None:
            for dl in _diff_group(display_group):
                yield dl
        while True:
            if args.interactive and apply_all is None:
                response = (yield Prompt("Apply this change? [ynadjk?]"))
                if not response:
                    continue
                response = response.lower()
            if not args.interactive or response.startswith('y') or apply_all == True:
                for old_line, new_line, line_or_conflict in display_group:
                    if isinstance(line_or_conflict, tuple):
                        preserve_changes[(old_line, new_line)] = line_or_conflict
                display_groups.pop(index)
                break
            elif response.startswith('n') or apply_all == False:
                for old_line, new_line, line_or_conflict in display_group:
                    if isinstance(line_or_conflict, tuple):
                        discard_changes[(old_line, new_line)] = line_or_conflict
                display_groups.pop(index)
                break
            elif response == 'j':
                index = (index + 1) % len(display_groups)
                break
            elif response == 'k':
                index = (index - 1) % len(display_groups)
                break
            elif response == 'a':
                index = 0
                apply_all = True
                break
            elif response == 'd':
                index = 0
                apply_all = False
                break
            elif response == '?':
                for l in apply.__doc__.split('\n')[-8:-1]:
                    yield l.strip()

    if not preserve_changes:
        yield "No changes in '%s' to apply." % os.path.relpath(changed_path)
        return

    # Build the changed file to be applied
    changes_to_apply = []

    i = 0
    old_line = 0
    new_line = 0
    while i < len(diff):
        line_or_conflict = diff[i]
        if isinstance(line_or_conflict, tuple):
            old, new = line_or_conflict
            if (old_line, new_line) in preserve_changes:
                changes_to_apply.extend(new)
            elif (old_line, new_line) in discard_changes:
                changes_to_apply.extend(old)
            else: # pragma: no cover
                raise Exception("Catastrophic error in selecting diff chunks. Please report a bug.")
            old_line += len(old)
            new_line += len(new)
            i += 1
        else:
            old_line += 1
            new_line += 1
            i += 1
            changes_to_apply.append(line_or_conflict)

    # Apply the changes across other files
    current_revision = changed_revision = 3
    weave.add_revision(changed_revision, changes_to_apply, [1])

    for other_path in _iterate_over_files(args.TARGET_FILENAME, config):
        other_key = os.path.relpath(other_path, config.root)
        if other_path == changed_path:
            continue # don't try to apply changes to ourself
        current_revision += 1
        weave.add_revision(current_revision, _smart_open(other_path, 'r').readlines(), [])
        merge_result = weave.cherry_pick(changed_revision, current_revision) # Can I apply changes in changed_revision onto this other file?
        if len(merge_result) == 1 and isinstance(merge_result[0], tuple):
            # total conflict, skip
            yield "Changes in '%s' cannot apply to '%s', skipping" % (os.path.relpath(changed_path), os.path.relpath(other_path))
        elif tuple in set(type(mr) for mr in merge_result):
            # some conflicts exist
            with _smart_open(other_path, 'w') as other_file:
                for line_or_conflict in merge_result:
                    if isinstance(line_or_conflict, tuple):
                        other_file.write('>'*7 + '\n')
                        for line in line_or_conflict[0]:
                            other_file.write(line)
                        other_file.write('='*7 + '\n')
                        for line in line_or_conflict[1]:
                            other_file.write(line)
                        other_file.write('>'*7 + '\n')
                    else:
                        other_file.write(line_or_conflict)
            yield "Conflict merging '%s' into '%s'" % (os.path.relpath(changed_path), os.path.relpath(other_path))
        else:
            # Merge is clean:
            with _smart_open(other_path, 'w') as other_file:
                other_file.writelines(merge_result)
            yield "Changes in '%s' applied cleanly to '%s'" % (os.path.relpath(changed_path), os.path.relpath(other_path))



# This code is based on code from
#   http://revctrl.org/PreciseCodevilleMerge?action=AttachFile
# That code was in turn based on BSD-licensed code from the Codeville distributed version control system
# -*- coding: utf-8
from __future__ import unicode_literals
try:
    xrange
except NameError: # Python3 compatibility
    xrange = range

from bisect import bisect

def unique_lcs(a, b):
    # set index[line in a] = position of line in a unless
    # unless a is a duplicate, in which case it's set to None
    index = {}
    for i in xrange(len(a)):
        line = a[i]
        if line in index:
            index[line] = None
        else:
            index[line]= i
    # make btoa[i] = position of line i in a, unless
    # that line doesn't occur exactly once in both,
    # in which case it's set to None
    btoa = [None] * len(b)
    index2 = {}
    for pos, line in enumerate(b):
        next = index.get(line)
        if next is not None:
            if line in index2:
                # unset the previous mapping, which we now know to
                # be invalid because the line isn't unique
                btoa[index2[line]] = None
                del index[line]
            else:
                index2[line] = pos
                btoa[pos] = next
    # this is the Patience sorting algorithm
    # see http://en.wikipedia.org/wiki/Patience_sorting
    backpointers = [None] * len(b)
    stacks = []
    lasts = []
    k = 0
    for bpos, apos in enumerate(btoa):
        if apos is None:
            continue
        # as an optimization, check if the next line comes at the end,
        # because it usually does
        if stacks and stacks[-1] < apos:
            k = len(stacks)
        # as an optimization, check if the next line comes right after
        # the previous line, because usually it does
        elif stacks and stacks[k] < apos and (k == len(stacks) - 1 or stacks[k+1] > apos):
            k += 1
        else:
            k = bisect(stacks, apos)
        if k > 0:
            backpointers[bpos] = lasts[k-1]
        if k < len(stacks):
            stacks[k] = apos
            lasts[k] = bpos
        else:
            stacks.append(apos)
            lasts.append(bpos)
    if len(lasts) == 0:
        return []
    result = []
    k = lasts[-1]
    while k is not None:
        result.append((btoa[k], k))
        k = backpointers[k]
    result.reverse()
    return result

def recurse_matches(a, b, ahi, bhi, answer, maxrecursion):
    oldlen = len(answer)
    if maxrecursion < 0:  # pragma: no cover
        # this will never happen normally, this check is to prevent DOS attacks
        return
    oldlength = len(answer)
    if len(answer) == 0:
        alo, blo = 0, 0
    else:
        alo, blo = answer[-1]
        alo += 1
        blo += 1
    if alo == ahi or blo == bhi:
        return
    for apos, bpos in unique_lcs(a[alo:ahi], b[blo:bhi]):
        # recurse between lines which are unique in each file and match
        apos += alo
        bpos += blo
        recurse_matches(a, b, apos, bpos, answer, maxrecursion - 1)
        answer.append((apos, bpos))
    if len(answer) > oldlength:
        # find matches between the last match and the end
        recurse_matches(a, b, ahi, bhi, answer, maxrecursion - 1)
    elif a[alo] == b[blo]:
        # find matching lines at the very beginning
        while alo < ahi and blo < bhi and a[alo] == b[blo]:
            answer.append((alo, blo))
            alo += 1
            blo += 1
        recurse_matches(a, b, ahi, bhi, answer, maxrecursion - 1)
    elif a[ahi - 1] == b[bhi - 1]:
        # find matching lines at the very end
        nahi = ahi - 1
        nbhi = bhi - 1
        while nahi > alo and nbhi > blo and a[nahi - 1] == b[nbhi - 1]:
            nahi -= 1
            nbhi -= 1
        recurse_matches(a, b, nahi, nbhi, answer, maxrecursion - 1)
        for i in xrange(ahi - nahi):
            answer.append((nahi + i, nbhi + i))

class Weave(object):
    def __init__(self):
        # [(lineid, line)]
        self.weave = []
        # {revid: [parent]}
        self.parents = {}
        # {revid: [((lineid1, lineid2), state)]}
        # states are integers
        # each edge's state starts at 0, then goes to 1, 2, etc.
        # odd states are when both lines are present and no lines are
        # between them, even otherwise
        # edges at the beginning and end are denoted by (None, lineid) and
        # (lineid, None) respectively. If the file is empty then (None, None)
        # is used.
        # the merge between two states is the greater of the two values
        self.newedgestates = {}

    def add_revision(self, revid, lines, parents):
        assert revid not in self.parents
        for p in parents:
            assert p in self.parents
        self.parents[revid] = [i for i in parents]

        # match against living lines
        # require that a line be part of living edges on either
        # side to be part of a living line,
        # to avoid including something in 'living lines' just
        # because a deletion happened next to it
        alivepre = set()
        alivepost = set()
        for ((ida, idb), state) in self._make_vals(revid).items():
            if state & 1 == 1:
                if ida is not None:
                    alivepre.add(ida)
                if idb is not None:
                    alivepost.add(idb)
        living = alivepre.intersection(alivepost)
        mapping = []
        livinglines = []
        for (pos, (lineid, line)) in enumerate(self.weave):
            if lineid in living:
                mapping.append(pos)
                livinglines.append(line)
        matches2 = []
        recurse_matches(lines, livinglines, len(lines), len(livinglines), matches2, 10)

        # match against the whole weave
        matches = []
        lines2 = [line for (lineid, line) in self.weave]
        for p, q in matches2:
            recurse_matches(lines, lines2, p, mapping[q], matches, 10)
            matches.append((p, mapping[q]))
        recurse_matches(lines, lines2, len(lines), len(lines2), matches, 10)

        # build a new weave
        alledges = set()
        for i in self.newedgestates.values():
            for (edge, state) in i:
                alledges.add(edge)
        newweave = []
        revpos = -1
        weavepos = -1
        matches.append((len(lines), len(lines2)))
        currentlines = []
        for a, b in matches:
            # take a guess as to whether it's better to put
            # extant lines before or after new lines
            hit = True
            if weavepos != -1 and weavepos + 1 != len(self.weave):
                hit = (self.weave[weavepos][0],
                    self.weave[weavepos + 1][0]) in alledges
            if hit:
                # add current weave lines to the new weave
                newweave.extend(self.weave[weavepos + 1:b])
            # add lines which have never appeared before to the weave
            for i in xrange(revpos + 1, a):
                lineid = (revid, i)
                currentlines.append(lineid)
                newweave.append((lineid, lines[i]))
            if not hit:
                # add current weave lines to the new weave
                newweave.extend(self.weave[weavepos + 1:b])
            if b != len(lines2):
                newweave.append(self.weave[b])
                currentlines.append(self.weave[b][0])
            revpos = a
            weavepos = b
        self.weave = newweave
        # calculate which lines had their states changed in this revision
        currentedges = set()
        if len(currentlines) > 0:
            for i in xrange(len(currentlines) - 1):
                currentedges.add((currentlines[i], currentlines[i+1]))
            currentedges.add((None, currentlines[0]))
            currentedges.add((currentlines[-1], None))
        else:
            currentedges.add((None, None))
        newedgevals = []
        vals = self._make_vals(revid)
        for edge in currentedges:
            if edge not in vals:
                newedgevals.append((edge, 1))
        for edge, state in vals.items():
            if (state & 1 == 1) != (edge in currentedges):
                newedgevals.append((edge, state + 1))
        if len(newedgevals) > 0:
            self.newedgestates[revid] = newedgevals

    def _make_vals(self, revid):
        # return {lineid: state} for the given revision
        unused = [revid]
        s = set()
        while unused:
            nextrev = unused.pop()
            if nextrev not in s:
                unused.extend(self.parents[nextrev])
                s.add(nextrev)
        v = {}
        for n in s:
            for p, q in self.newedgestates.get(n, []):
                v[p] = max(v.get(p, 0), q)
        return v

    def _lineids(self, vals):
        # return set of lineids of lines alive in output of _make_vals
        lineids = set()
        for (ida, idb), state in vals.items():
            if state & 1 == 1:
                if ida is not None:
                    lineids.add(ida)
                if idb is not None:
                    lineids.add(idb)
        return lineids

    def retrieve_revision(self, revid):
        # returns a list of strings
        ids = self._lineids(self._make_vals(revid))
        return [line for (lineid, line) in self.weave if lineid in ids]

    def cherry_pick(self, reva, revb):
        # pulls just the change in reva (without history) into revb
        v = {}
        for p, q in self.newedgestates.get(reva, []):
            v[p] = max(v.get(p, 0), q)

        alines = self._lineids(self._make_vals(reva))
        return self.merge(reva, revb, edgesa=v, alines=alines)

    def merge(self, reva, revb, edgesa=None, alines=None, edgesb=None, blines=None):
        # returns [line]
        # non-conflict lines are strings, conflict sections are
        # ([linesa], [linesb])
        if edgesa is None:
            edgesa = self._make_vals(reva)
        if alines is None:
            alines = self._lineids(edgesa)
        if edgesb is None:
            edgesb = self._make_vals(revb)
        if blines is None:
            blines = self._lineids(edgesb)
        lastalineid = None
        lastblineid = None
        awins = False
        bwins = False
        apartial = []
        bpartial = []
        result = []
        for (lineid, line) in (self.weave + [(None, None)]):
            if lineid is None or lineid in alines:
                edge = (lastalineid, lineid)
                aval = edgesa.get(edge, 0)
                bval = edgesb.get(edge, 0)
                if aval > bval:
                    awins = True
                if bval > aval:
                    bwins = True
                lastalineid = lineid
            if lineid is None or lineid in blines:
                edge = (lastblineid, lineid)
                aval = edgesa.get(edge, 0)
                bval = edgesb.get(edge, 0)
                if aval > bval:
                    awins = True
                if bval > aval:
                    bwins = True
                lastblineid = lineid
            if lineid is None or (lineid in alines and lineid in blines):
                #if not (awins ^ bwins):
                if awins and bwins:
                    result.append((apartial, bpartial))
                elif awins:
                    result.extend(apartial)
                elif bwins:
                    result.extend(bpartial)
                awins = False
                bwins = False
                apartial = []
                bpartial = []
                if line is not None:
                    result.append(line)
            else:
                if lineid in alines:
                    apartial.append(line)
                if lineid in blines:
                    bpartial.append(line)
        return result



#!/usr/bin/env python
# -*- coding: utf-8
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

install_requires = []
if sys.version_info < (2, 7):
    install_requires.append('argparse')

import fragments

setup(
    name='fragments',
    version='.'.join(map(str, fragments.__version__)),
    description='tool for diff-based templating / fragmentation control',
    long_description=open('README.md').read(),
    author='Matt Chisholm',
    author_email='matt-fragments@theory.org',
    url='https://github.com/glyphobet/fragments',
    packages= ['fragments',],
    install_requires=install_requires,
    entry_points = {
        'console_scripts': [
            'fragments = fragments.commands:_main',
        ],
    },
    data_files=[('share/fragments', ['README.md', 'bash_completion', 'LICENSE.txt'])],
    license='BSD License',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
    ),
)
